module PickStrategy(pickStrategy) where

import Strategy
import Robot
import GameState
import List
import Commands
import Board

-- Strategy for picking things up

binPack :: Int -> [Package] -> [Package]
binPack i packs = simplePack i packs

simplePack i [] = []
simplePack i (pack:packs) | pw pack <= i = pack : simplePack (i - pw pack) packs
			  | otherwise    = simplePack i packs

sameDestSort pos packs
    = sortBy comparePos packs
  where comparePos a b = compare (dist (packPos a) pos) (dist (packPos b) pos)

pickStrategy :: Rob (Maybe Command)
pickStrategy
    = inspect (stateR & botRadarR 1) $ \ (s,closeBots) ->
      let hotPacks = myPacks s ++ pickablePacks s
	  wantedPacks = binPack (wholeCapacity s)
                                (sameDestSort (myPos s) hotPacks)
	  superfluous = myPacks s \\ wantedPacks
      in if totalWeight wantedPacks > totalWeight (myPacks s)
         then if not (null superfluous) && null closeBots
	      then Just (C 1 (Drop (map pid superfluous)))
	      else Just (C 1 (Pick (map pid (wantedPacks \\ myPacks s))))
	 else Nothing

totalWeight :: [Package] -> Int
totalWeight = sum . map pw
