module Robot where

import Commands
import Board
import GameState
import Memo

infixl 6 ==>
infixl 5 .|.
infixr 5 &
infixr 4 `switch`

---------------------------------------------------------------------------
-- Rob type

data Rob a
  = Result a (Rob a)
  | State (StateR -> Rob a)
  | Const a

instance Functor Rob where
  fmap = lift1

type StateR = (State, Extra)

-- extra, memoized state information

type Extra = Pos -> Maybe (Int,[Direction])

stateExtra :: State -> Extra
stateExtra s = memo (shortestPath (board s) (myPos s))

instance Memo Pos where
  memo f = memo (f . toPos) . fromPos
   where
    toPos   (x,y)     = Pos x y
    fromPos (Pos x y) = (x,y)

---------------------------------------------------------------------------
-- primitive functions

constR :: a -> Rob a
constR a = Const a

lift1 :: (a -> b) -> Rob a -> Rob b
lift1 f (Const a)    = Const (f a)
lift1 f (State k)    = State (lift1 f . k)
lift1 f (Result a r) = Result (f a) (lift1 f r)

lift2 :: (a -> b -> c) -> Rob a -> Rob b -> Rob c
lift2 f (Const a)     r2            = lift1 (f a) r2
lift2 f r1            (Const b)     = lift1 (`f` b) r1
lift2 f (State k1)    (State k2)    = State (\s -> lift2 f (k1 s) (k2 s))
lift2 f (State k1)    r2            = State (\s -> lift2 f (k1 s) r2)
lift2 f r1            (State k2)    = State (\s -> lift2 f r1     (k2 s))
lift2 f (Result a r1) (Result b r2) = Result (f a b) (lift2 f r1 r2)

change :: Rob (a, Maybe (Rob a)) -> Rob a
change (Result (a,Nothing) r) = Result a (change r)
change (Result (_,Just r') _) = r'
change (Const (a,Nothing))    = Const a
change (Const (_,Just r'))    = r'
change (State k)              = State (\s -> change (k s))

stateR :: Rob State
stateR = State (\(s,_) -> Result s stateR)

extraR :: Rob Extra
extraR = State (\(_,x) -> Result x extraR)

getState :: (State -> Rob a) -> Rob a
getState f = State (\(s,_) -> f s)

shift :: a -> Rob a -> Rob a
shift a r = Result a r

-- some events

event :: Rob Bool -> Rob (Maybe ())
event = lift1 (\b -> if b then Just () else Nothing)

one :: a -> Rob (Maybe a)
one a = Result (Just a) (Const Nothing)

after :: Int -> Rob (Maybe ())
after 0 = one ()
after n = Result Nothing (after (n-1))

every :: Int -> Rob (Maybe ())
every n = let loop 0 = Result (Just ()) loopn
              loop k = Result Nothing (loop (k-1))
              loopn  = loop n
           in loopn

---------------------------------------------------------------------------
-- derived functions

switch :: Rob a -> Rob (Maybe (Rob a)) -> Rob a
switch r rm = change (r & rm)

switcher :: Rob (a,Maybe b) -> (b -> Rob a) -> Rob a
switcher r f = change (fmap (\(a,mb) -> (a,f `fmap` mb)) r)

switching :: Rob a -> Rob (Maybe (Rob a)) -> Rob a
switching r mr = r `switch` mr ==> \r' -> switching r' mr

(==>) :: Rob (Maybe a) -> (a -> b) -> Rob (Maybe b)
evt ==> f = lift1 (fmap f) evt

(.=>) :: Rob (Maybe a) -> b -> Rob (Maybe b)
evt .=> b = lift1 (fmap (const b)) evt

(.|.) :: Rob (Maybe a) -> Rob (Maybe a) -> Rob (Maybe a)
(.|.) = lift2 (\a b -> maybe b Just a)

lift0 :: a -> Rob a
lift0 = constR

inspect :: Rob a -> (a -> b) -> Rob b
inspect r f = lift1 f r

(&) :: Rob a -> Rob b -> Rob (a,b)
(&) = lift2 (,)

hold :: a -> Rob (Maybe a) -> Rob a
hold a newR = constR a `switch` newR ==> \a' -> hold a' newR

---------------------------------------------------------------------------
-- run function

run :: State -> Rob Command -> (Command, Rob Command)
run s = run'
 where
  sx = (s, stateExtra s)
  
  run' (Result m r) = (m, r)
  run' r@(Const m)  = (m, r)
  run' (State k)    = run' (k sx)

---------------------------------------------------------------------------
-- the end.
