module Strategy where

import Robot
import Commands
import GameState
import Board
import Data.FiniteMap
import Random

---------------------------------------------------------------------------
-- commands

command :: Action -> Command
command act = C 1 act

bidding :: Integer -> Command -> Command
bidding n (C _ act) = C n act

stay :: Command
stay = command nop

walk :: Direction -> Command
walk dir = command (Move dir)

get :: [PackageID] -> Command
get ids = command (Pick ids)

put :: [PackageID] -> Command
put ids = command (Drop ids)

---------------------------------------------------------------------------
-- state functions

shortestR :: Rob (Pos -> Maybe (Int,[Direction]))
shortestR = extraR

---------------------------------------------------------------------------
-- basic robot behaviors

stayR :: Rob Command
stayR = constR stay

walkToR :: Rob (Maybe Pos) -> Rob Command
walkToR posR =
  inspect (posR & shortestR) $ \(mpos, short) ->
    case do pos       <- mpos
            (n,dir:_) <- short pos
            return dir of
      Nothing  -> stay
      Just dir -> walk dir

keepWalkingToR :: Rob (Maybe Pos) -> Rob Command
keepWalkingToR posR = walkToR (hold Nothing (Just `fmap` posR))

randomWalkR :: Rob Command
randomWalkR =
  getState (\s ->
    walkToR (constR (Just (randomPos s))) `switch`
      after 5 .=> randomWalkR
  )
 where
  randomPos s =
    head [ pos
         | r <- rnds (randomSeed s)
         , let (r1,r2) = split r
               (i,_)   = randomR (1,x) r1
               (j,_)   = randomR (1,y) r2
               pos     = Pos i j
         , board s ? pos == Empty
         ]
   where
    Dim x y = dimensions (board s)
    rnds r  = let (r1,r2) = split r in r1 : rnds r2

---------------------------------------------------------------------------
-- some strategies

-- a radar for finding nearby bots

botRadarR :: Int -> Rob [(RobotID,Pos)]
botRadarR i =
  inspect stateR $ \s ->
    [ (rid,pos)
    | (rid,pos) <- fmToList (otherBots s)
    , pos `dist` myPos s <= i
    ]

---------------------------------------------------------------------------
-- the end.
